
// Define a Emptyable protocol which has a isEmpty property so we can use it to check whether a value is considered empty
protocol Emptyable {
    var isEmpty: Bool { get }
}

//Extend the Optional protocol and add our function to provide a default value if the Optional is nil or empty
extension Optional where Wrapped: Emptyable {
    func orWhenNilOrEmpty(_ defaultValue: Wrapped) -> Wrapped {
        switch(self) {
        case .none:
            return defaultValue
        case .some(let value) where value.isEmpty:
            return defaultValue
        case .some(let value):
            return value
        }
    }
}

//Extend the type where we want to use the previously created extension of Optional
extension String: Emptyable {}


var optionalString: String? = nil
optionalString.orWhenNilOrEmpty("Something")

optionalString = ""
optionalString.orWhenNilOrEmpty("Something")

optionalString = "Hello World"
optionalString.orWhenNilOrEmpty("Something")


// Of course this works out of the box on every type which already has an isEmpty property, like arrays
extension Array: Emptyable {}

var optionalArray: [Any]? = nil
optionalArray.orWhenNilOrEmpty(["Hello", "World"])

optionalArray = []
optionalArray.orWhenNilOrEmpty(["Hello", "World"])

optionalArray = ["This", "is", "so", "great"]
optionalArray.orWhenNilOrEmpty(["Hello", "World"])


// We can also implement our own logic if the type has no isEmpty by default. For example when using custom types
struct MyStruct: Emptyable, CustomStringConvertible {
    var text: String
    var number: Int
    var isEmpty: Bool { return text.isEmpty && number == 0 }
    var description: String { return "\(text) \(number)" }
    init(text: String, number: Int) {
        self.text = text
        self.number = number
    }
}

var myOptional: MyStruct? = nil
var myMandatory: MyStruct = MyStruct(text:"Hello World", number: 123)
myOptional.orWhenNilOrEmpty(myMandatory)


myOptional = MyStruct(text: "", number: 0)
myOptional.orWhenNilOrEmpty(myMandatory)


myOptional = MyStruct(text: "I am number", number: 1)
myOptional.orWhenNilOrEmpty(myMandatory)





// And no lets get really stupid by creating our own operator to call this function
infix operator ???: NilCoalescingPrecedence

extension Optional where Wrapped: Emptyable {
    static func ???(left: Wrapped?, right: Wrapped) -> Wrapped {
        return left.orWhenNilOrEmpty(right)
    }
}

// Time to try it out
let noString: String? = nil
noString ??? "Hello World"
noString ?? "" ??? "Hello World"
noString ?? "Okay" ??? "Hello World"


let noStruct: MyStruct? = nil
noStruct ??? MyStruct(text: "Not Empty", number: 1)
noStruct ?? myOptional ??? MyStruct(text: "Not Empty", number: 1)



